from __future__ import annotations

from collections.abc import Coroutine
from contextvars import ContextVar

from tekore._sender import Client, Request, Response, Sender

prefix = "https://api.spotify.com/v1/"


def build_url(url: str) -> str:
    """Attach API address to endpoint."""
    if not url.startswith("http"):
        url = prefix + url
    return url


def parse_url_params(params: dict | None) -> dict | None:
    """Generate parameter dict and filter Nones."""
    params = params or {}
    return {k: v for k, v in params.items() if v is not None} or None


class SpotifyBase(Client):
    """Base client with options and utility functions."""

    _token_cv = ContextVar("_token_cv")
    _max_limits_on_cv = ContextVar("_max_limits_on_cv")
    _chunked_on_cv = ContextVar("_chunked_on_cv")

    def __init__(
        self,
        token=None,
        sender: Sender | None = None,
        asynchronous: bool | None = None,
        max_limits_on: bool = False,
        chunked_on: bool = False,
    ) -> None:
        # Docstring in the main client
        super().__init__(sender, asynchronous)
        self._token = token
        self._max_limits_on = max_limits_on
        self._chunked_on = chunked_on

    @property
    def token(self):
        """Token getter."""
        return self._token_cv.get(self._token)

    @token.setter
    def token(self, value) -> None:
        try:
            self._token_cv.get()
        except LookupError:
            self._token = value
        else:
            self._token_cv.set(value)

    @property
    def max_limits_on(self) -> bool:
        """Max limits getter."""
        return self._max_limits_on_cv.get(self._max_limits_on)

    @max_limits_on.setter
    def max_limits_on(self, value: bool) -> None:
        try:
            self._max_limits_on_cv.get()
        except LookupError:
            self._max_limits_on = value
        else:
            self._max_limits_on_cv.set(value)

    @property
    def chunked_on(self) -> bool:
        """Chunked getter."""
        return self._chunked_on_cv.get(self._chunked_on)

    @chunked_on.setter
    def chunked_on(self, value: bool) -> None:
        try:
            self._chunked_on_cv.get()
        except LookupError:
            self._chunked_on = value
        else:
            self._chunked_on_cv.set(value)

    def __repr__(self) -> str:
        options = [
            f"token={self.token!r}",
            f"max_limits_on={self.max_limits_on}",
            f"chunked_on={self.chunked_on}",
            f"sender={self.sender!r}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    def _create_headers(self, content_type: str = "application/json"):
        return {"Authorization": f"Bearer {self.token!s}", "Content-Type": content_type}

    def send(self, request: Request) -> Response | Coroutine[None, None, Response]:
        """
        Build request url and headers, and send with underlying sender.

        Exposed to easily send arbitrary requests,
        for custom behavior in some endpoint e.g. for a subclass.
        It may also come in handy if a bugfix or a feature is not implemented
        in a timely manner, or in debugging related to the client or Web API.
        """
        request.url = build_url(request.url)
        headers = self._create_headers()
        if request.headers is not None:
            headers.update(request.headers)
        request.headers = headers
        return self.sender.send(request)

    @staticmethod
    def _request(
        method: str, url: str, payload: dict | None = None, params: dict | None = None
    ) -> tuple[Request, tuple]:
        return (
            Request(
                method=method, url=url, params=parse_url_params(params), json=payload
            ),
            (),
        )

    def _get(self, url: str, payload: dict | None = None, **params):
        return self._request("GET", url, payload=payload, params=params)

    def _post(self, url: str, payload: dict | None = None, **params):
        return self._request("POST", url, payload=payload, params=params)

    def _delete(self, url: str, payload: dict | None = None, **params):
        return self._request("DELETE", url, payload=payload, params=params)

    def _put(self, url: str, payload: dict | None = None, **params):
        return self._request("PUT", url, payload=payload, params=params)



from enum import Enum


class scope(Enum):  # noqa: N801
    """
    User access token privileges.

    The string representation of a member is its enum value.

    .. code:: python

       s = tk.scope.user_read_email
       print(s)  # -> 'user-read-email'

    Also provides three scopes that are a combination of others.

    .. code:: python

        tk.scope.read: Scope = ...            # All read scopes
        tk.scope.write: Scope = ...           # All write scopes
        tk.scope.every: Scope = read + write  # All available scopes

    .. note::

        :attr:`app_remote_control` and :attr:`streaming` are only used outside
        of the Web API, and are not included in the premade scope combinations.

    Addition and subtraction from both sides is supported
    but delegated to :class:`Scope` and always returns a :class:`Scope`.
    """

    user_read_email = "user-read-email"
    user_read_private = "user-read-private"
    user_top_read = "user-top-read"
    user_read_recently_played = "user-read-recently-played"

    user_follow_read = "user-follow-read"
    user_follow_modify = "user-follow-modify"
    user_library_read = "user-library-read"
    user_library_modify = "user-library-modify"

    user_read_currently_playing = "user-read-currently-playing"
    user_read_playback_state = "user-read-playback-state"
    user_read_playback_position = "user-read-playback-position"
    user_modify_playback_state = "user-modify-playback-state"

    playlist_modify_public = "playlist-modify-public"
    playlist_read_collaborative = "playlist-read-collaborative"
    playlist_read_private = "playlist-read-private"
    playlist_modify_private = "playlist-modify-private"

    ugc_image_upload = "ugc-image-upload"

    app_remote_control = "app-remote-control"
    streaming = "streaming"

    def __str__(self) -> str:
        """Enum value."""
        return self.value

    def __add__(self, other) -> "Scope":
        """Combine to a set of scopes."""
        return Scope(self) + other

    def __radd__(self, other) -> "Scope":
        """Combine to a set of scopes."""
        return other + Scope(self)

    def __sub__(self, other) -> "Scope":
        """Remove scope from another."""
        return Scope(self) - other

    def __rsub__(self, other) -> "Scope":
        """Remove scope from another."""
        return other - Scope(self)


class Scope(frozenset):
    """
    Set of :class:`scopes <scope>` for a token.

    Instantiated with an unpacked list of strings or :class:`scopes <scope>`.

    .. code:: python

        bruce = tk.Scope(*tk.scope)
        sally = tk.Scope(tk.scope.user_read_email, 'ugc-image-upload')
        elise = tk.Scope(*sally, *timmy, tk.scope.user_follow_modify)

    Also supports flexible addition and subtraction from both sides
    with strings, :class:`scopes <scope>` and other :class:`Scope` objects.
    Addition is a set-like union, subtraction is a set-like relative complement.
    If any operation is unsuccessful, :class:`NotImplementedError` is raised.

    .. code:: python

        waldo = tk.scopes.user_follow_read + sally + elise - 'user-read-email'

    The string representation of a :class:`Scope` is a sorted,
    space-separated concatenation of its members.

    .. code:: python

       s = tk.Scope('b', 'c', 'a')
       print(s)  # -> 'a b c'
    """

    def __new__(cls, *members) -> "Scope":
        """
        Construct a new set of scopes.

        Parameters
        ----------
        members
            unpacked list of members of the new scope
        """
        return super().__new__(cls, [str(m) for m in members])

    def __repr__(self) -> str:
        """Readable representation."""
        members = "'" + "', '".join(sorted(self)) + "'"
        return f"Scope({members})"

    def __str__(self) -> str:
        """Join members with spaces."""
        return " ".join(sorted(self))

    def __add__(self, other) -> "Scope":
        """Combine two sets of scopes."""
        if isinstance(other, (str, scope)):
            other = {str(other)}
        elif not isinstance(other, Scope):
            e = f"Addition not defined for {type(self)} and {type(other)}!"
            raise NotImplementedError(e)
        return type(self)(*self.union(other))

    def __radd__(self, other) -> "Scope":
        """Combine two sets of scopes."""
        return self + other

    def __sub__(self, other) -> "Scope":
        """Remove scopes from a set."""
        if isinstance(other, (str, scope)):
            other = {str(other)}
        elif not isinstance(other, Scope):
            e = f"Difference not defined for {type(self)} and {type(other)}!"
            raise NotImplementedError(e)
        return type(self)(*self.difference(other))

    def __rsub__(self, other) -> "Scope":
        """Remove scopes from a set."""
        if not isinstance(other, (str, scope)):
            e = f"Difference not defined for {type(other)} and {type(self)}!"
            raise NotImplementedError(e)
        return Scope(other) - self


scope.read = Scope(
    scope.user_read_email,
    scope.user_read_private,
    scope.user_top_read,
    scope.user_read_recently_played,
    scope.user_follow_read,
    scope.user_library_read,
    scope.user_read_currently_playing,
    scope.user_read_playback_state,
    scope.user_read_playback_position,
    scope.playlist_read_collaborative,
    scope.playlist_read_private,
)
scope.write = Scope(
    scope.user_follow_modify,
    scope.user_library_modify,
    scope.user_modify_playback_state,
    scope.playlist_modify_public,
    scope.playlist_modify_private,
    scope.ugc_image_upload,
)
scope.every = scope.read + scope.write



from __future__ import annotations

from typing import Literal

from pydantic import Field

from .serialise import Model


class LocalItem(Model):
    """Base for local items."""

    id: None
    href: None
    name: str
    type: str


class LocalAlbum(LocalItem):
    """Album of a locally saved track."""

    album_type: None
    artists: list[None]
    available_markets: list[None] = Field(default_factory=list)
    external_urls: dict
    images: list[None]
    release_date: None
    release_date_precision: None
    uri: None


class LocalArtist(LocalItem):
    """Artist of a locally saved track."""

    external_urls: dict
    uri: None


class LocalTrack(LocalItem):
    """
    Locally saved track.

    Locally saved track where most attributes are
    always None, empty, zero or False.
    :attr:`duration_ms` being an object is undocumented.
    """

    album: LocalAlbum
    artists: list[LocalArtist]
    available_markets: list[None] = Field(default_factory=list)
    disc_number: int
    duration_ms: int | dict
    explicit: bool
    external_ids: dict
    external_urls: dict
    is_local: Literal[True]
    popularity: int
    preview_url: None
    track_number: int
    uri: str



from .expiring import AccessToken, Credentials, Token
from .refreshing import RefreshingCredentials, RefreshingToken
from .scope import Scope, scope
from .util import (
    UserAuth,
    gen_state,
    parse_code_from_url,
    parse_state_from_url,
    prompt_for_pkce_token,
    prompt_for_user_token,
    refresh_pkce_token,
    refresh_user_token,
    request_client_token,
)



from httpx import codes

from tekore._sender import Request, Response, send_and_process
from tekore._sender.base import Sender

from .base import SpotifyBase


def is_short_link(url: str) -> bool:
    """Determine if URL is a Spotify short link."""
    return "spotify.link/" in url


def _process_short_link(request: Request, response: Response) -> str:
    if response.status_code == codes.TEMPORARY_REDIRECT:
        return response.headers["location"]
    return request.url


@send_and_process(_process_short_link)
def _send_short_link(_: Sender, request: tuple[Request, tuple]) -> str:
    return request


class SpotifyShortLink(SpotifyBase):
    """Spotify short link."""

    def follow_short_link(self, link: str) -> str:
        """
        Follow redirect of a short link to get the underlying resource URL.

        Safely also accept a direct link, request a redirect and return
        the original URL. Also use the underlying sender for an unauthenticated request.

        Returns
        -------
        url
            result of the short link redirect
        """
        request = self._request("HEAD", link)
        return _send_short_link(self.sender, request)



from tekore._client.base import SpotifyBase
from tekore._client.decor import scopes, send_and_process
from tekore._client.process import top_item


class SpotifyMarkets(SpotifyBase):
    """Markets API endpoints."""

    @scopes()
    @send_and_process(top_item("markets"))
    def markets(self) -> list[str]:
        """
        Get available market country codes.

        Returns
        -------
        list[str]
            available markets
        """
        return self._get("markets")



from .base import Request, Response, Sender
from .client import Client, SenderConflictWarning, send_and_process
from .concrete import AsyncSender, SyncSender
from .error import (
    BadGateway,
    BadRequest,
    ClientError,
    Forbidden,
    HTTPError,
    InternalServerError,
    NotFound,
    ServerError,
    ServiceUnavailable,
    TooManyRequests,
    Unauthorised,
)
from .extending import CachingSender, ExtendingSender, RetryingSender



from __future__ import annotations

from base64 import b64encode as _b64encode
from hashlib import sha256
from secrets import token_urlsafe
from urllib.parse import urlencode

from tekore._auth.scope import Scope
from tekore._sender import Client, Request, Sender, send_and_process

from .decor import parse_refreshed_token, parse_token
from .token import Token

OAUTH_AUTHORIZE_URL = "https://accounts.spotify.com/authorize"
OAUTH_TOKEN_URL = "https://accounts.spotify.com/api/token"


def b64encode(msg: str) -> str:
    """Encode a unicode string in base-64."""
    return _b64encode(msg.encode()).decode()


def b64urlencode(msg: bytes) -> str:
    """Encode bytes in url-safe base-64 alphabet."""
    encoded = _b64encode(msg).decode()
    stripped = encoded.split("=")[0]
    return stripped.replace("+", "-").replace("/", "_")


class Credentials(Client):
    """
    Client for retrieving access tokens.

    Parameters
    ----------
    client_id
        client id
    client_secret
        client secret, not required for PKCE user authorisation
    redirect_uri
        whitelisted redirect URI, required for user authorisation
    sender
        request sender
    asynchronous
        synchronicity requirement
    """

    def __init__(
        self,
        client_id: str,
        client_secret: str | None = None,
        redirect_uri: str | None = None,
        sender: Sender | None = None,
        asynchronous: bool | None = None,
    ) -> None:
        super().__init__(sender, asynchronous)
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

    def __repr__(self) -> str:
        options = [
            f"client_id={self.client_id!r}",
            f"client_secret={self.client_secret!r}",
            f"redirect_uri={self.redirect_uri!r}",
            f"sender={self.sender!r}",
        ]
        return type(self).__name__ + "(" + ", ".join(options) + ")"

    def _token_request(self, payload: dict, *, auth: bool) -> Request:
        if auth:
            if self.client_secret is None:
                msg = f"A client secret is required! Got `{self.client_secret}`."
                raise ValueError(msg)
            token = b64encode(self.client_id + ":" + self.client_secret)
            headers = {"Authorization": f"Basic {token}"}
        else:
            headers = None

        return Request("POST", OAUTH_TOKEN_URL, data=payload, headers=headers)

    @send_and_process(parse_token(uses_pkce=False))
    def request_client_token(self) -> Token:
        """
        Request a client token.

        Returns
        -------
        Token
            client access token
        """
        payload = {"grant_type": "client_credentials"}
        return self._token_request(payload, auth=True), ()

    def _user_auth_payload(self, scope, state: str | None):
        payload = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "response_type": "code",
        }
        if isinstance(scope, list):
            scope = Scope(*scope)
        if scope is not None:
            payload["scope"] = str(scope)
        if state is not None:
            payload["state"] = state
        return payload

    def user_authorisation_url(
        self, scope=None, state: str | None = None, show_dialog: bool = False
    ) -> str:
        """
        Construct an authorisation URL.

        Step 1/2 in authorisation code flow.
        User should be redirected to the resulting URL for authorisation.
        Step 2/2: :meth:`request_user_token`.

        Parameters
        ----------
        scope
            token privileges, accepts a :class:`Scope`, a single :class:`scope`,
            a list of :class:`scopes <scope>` and strings for :class:`Scope`,
            or a space-separated list of scopes as a string
        state
            additional state
        show_dialog
            force login dialog even if previously authorised

        Returns
        -------
        str
            login URL
        """
        payload = self._user_auth_payload(scope, state)
        payload["show_dialog"] = str(show_dialog).lower()
        return OAUTH_AUTHORIZE_URL + "?" + urlencode(payload)

    @send_and_process(parse_token(uses_pkce=False))
    def request_user_token(self, code: str) -> Token:
        """
        Request a new user token.

        Step 2/2 in authorisation code flow.
        Code is provided as a URL parameter in the redirect URI
        after login in step 1: :meth:`user_authorisation_url`.

        Parameters
        ----------
        code
            code from redirect parameters

        Returns
        -------
        Token
            user access token
        """
        payload = {
            "code": code,
            "redirect_uri": self.redirect_uri,
            "grant_type": "authorization_code",
        }
        return self._token_request(payload, auth=True), ()

    @send_and_process(parse_refreshed_token(uses_pkce=False))
    def refresh_user_token(self, refresh_token: str) -> Token:
        """
        Request a refreshed user token.

        Parameters
        ----------
        refresh_token
            refresh token

        Returns
        -------
        Token
            refreshed user access token
        """
        payload = {"refresh_token": refresh_token, "grant_type": "refresh_token"}
        return self._token_request(payload, auth=True), (refresh_token,)

    def pkce_user_authorisation(
        self, scope=None, state: str | None = None, verifier_bytes: int = 32
    ) -> tuple[str, str]:
        """
        Construct authorisation URL and verifier.

        Step 1/2 in authorisation code flow with proof key for code exchange.
        The user should be redirected to the resulting URL for authorisation.
        The verifier is passed to :meth:`request_pkce_token` in step 2.

        Parameters
        ----------
        scope
            token privileges, accepts a :class:`Scope`, a single :class:`scope`,
            a list of :class:`scopes <scope>` and strings for :class:`Scope`,
            or a space-separated list of scopes as a string
        state
            additional state
        verifier_bytes
            number of bytes to generate PKCE verifier with, ``32 <= bytes <= 96``.
            The specified range of bytes generates the appropriate number of
            characters (43 - 128) after base-64 encoding, as required in RFC 7636.

        Returns
        -------
        tuple[str, str]
            authorisation URL and PKCE code verifier
        """
        if not (32 <= verifier_bytes <= 96):  # noqa: PLR2004
            msg = "Invalid number of verifier bytes!"
            raise AssertionError(msg)
        verifier = token_urlsafe(verifier_bytes)

        sha = sha256(verifier.encode())
        challenge = b64urlencode(sha.digest())

        payload = self._user_auth_payload(scope, state)
        payload["code_challenge"] = challenge
        payload["code_challenge_method"] = "S256"

        auth_url = OAUTH_AUTHORIZE_URL + "?" + urlencode(payload)
        return auth_url, verifier

    @send_and_process(parse_token(uses_pkce=True))
    def request_pkce_token(self, code: str, verifier: str) -> Token:
        """
        Request a new PKCE user token.

        Step 2/2 in authorisation code flow with proof key for code exchange.
        Code is provided as a URL parameter in the redirect URI
        after login in step 1: :meth:`pkce_user_authorisation`.

        Parameters
        ----------
        code
            code from redirect parameters
        verifier
            PKCE code verifier generated for authorisation URL

        Returns
        -------
        Token
            user access token
        """
        payload = {
            "client_id": self.client_id,
            "code": code,
            "code_verifier": verifier,
            "grant_type": "authorization_code",
            "redirect_uri": self.redirect_uri,
        }
        return self._token_request(payload, auth=False), ()

    @send_and_process(parse_refreshed_token(uses_pkce=True))
    def refresh_pkce_token(self, refresh_token: str) -> Token:
        """
        Request a refreshed PKCE user token.

        Parameters
        ----------
        refresh_token
            refresh token

        Returns
        -------
        Token
            refreshed user access token
        """
        payload = {
            "client_id": self.client_id,
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        }
        return self._token_request(payload, auth=False), (refresh_token,)

    def refresh(self, token: Token) -> Token:
        """
        Refresh an access token.

        Both client and user tokens are accepted and refreshed.
        The correct refreshing method is applied regardless if PKCE was used or not.
        For client tokens, a new token is returned.
        For user tokens, a refreshed token is returned.

        Parameters
        ----------
        token
            token to be refreshed

        Returns
        -------
        Token
            refreshed access token
        """
        if token.refresh_token is None:
            return self.request_client_token()
        if token.uses_pkce:
            return self.refresh_pkce_token(token.refresh_token)
        return self.refresh_user_token(token.refresh_token)



from __future__ import annotations

from tekore._auth import scope
from tekore._client.base import SpotifyBase
from tekore._client.decor import maximise_limit, scopes, send_and_process
from tekore._client.process import model_list, single
from tekore.model import (
    CurrentlyPlaying,
    CurrentlyPlayingContext,
    Device,
    PlayHistoryPaging,
    Queue,
)


class SpotifyPlayerView(SpotifyBase):
    """Player API endpoints that view state."""

    @scopes([scope.user_read_playback_state])
    @send_and_process(single(CurrentlyPlayingContext))
    def playback(
        self, market: str | None = None, tracks_only: bool = False
    ) -> CurrentlyPlayingContext:
        """
        Get information about user's current playback.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        tracks_only
            return only tracks in the currently playing item,
            if True, episodes have None as the currently playing item

        Returns
        -------
        CurrentlyPlayingContext
            information about current playback
        """
        additional_types = None if tracks_only else "episode"

        return self._get("me/player", market=market, additional_types=additional_types)

    @scopes(
        [scope.user_read_playback_state, scope.user_read_currently_playing],
        [scope.user_read_playback_state, scope.user_read_currently_playing],
    )
    @send_and_process(single(CurrentlyPlaying))
    def playback_currently_playing(
        self, market: str | None = None, tracks_only: bool = False
    ) -> CurrentlyPlaying:
        """
        Get user's currently playing track.

        Only one of the scopes above is required.

        Parameters
        ----------
        market
            an ISO 3166-1 alpha-2 country code or 'from_token'
        tracks_only
            return only tracks in the currently playing item,
            if True, episodes have None as the currently playing item

        Returns
        -------
        CurrentlyPlaying
            information about the current track playing
        """
        additional_types = None if tracks_only else "episode"

        return self._get(
            "me/player/currently-playing",
            market=market,
            additional_types=additional_types,
        )

    @scopes([scope.user_read_recently_played])
    @send_and_process(single(PlayHistoryPaging))
    @maximise_limit(50)
    def playback_recently_played(
        self, limit: int = 20, after: int | None = None, before: int | None = None
    ) -> PlayHistoryPaging:
        """
        Get tracks from the current user's recently played tracks.

        Only ``after`` or ``before`` should be specified at one time.

        Parameters
        ----------
        limit
            the number of items to return (1..50)
        after
            a unix timestamp in milliseconds, must not be specified with 'before'
        before
            a unix timestamp in milliseconds, must not be specified with 'after'
        """
        return self._get(
            "me/player/recently-played", limit=limit, after=after, before=before
        )

    @scopes([scope.user_read_playback_state])
    @send_and_process(model_list(Device, "devices"))
    def playback_devices(self) -> list[Device]:
        """Get a user's available devices."""
        return self._get("me/player/devices")

    @scopes([scope.user_read_playback_state])
    @send_and_process(single(Queue))
    def playback_queue(self) -> Queue:
        """
        Get items in a user's queue.

        The result also include items in the current playback context
        even though they are not explicitly in the queue.
        The number of items in the queue is inconsistent,
        but from manual experimentation at least two items are returned.
        """
        return self._get("me/player/queue")



from .base import Identifiable
from .member import Image
from .paging import OffsetPaging


class Category(Identifiable):
    """Spotify tag category."""

    href: str
    icons: list[Image]
    name: str


class CategoryPaging(OffsetPaging):
    """Paging of categories."""

    items: list[Category]



from __future__ import annotations

import operator
from collections.abc import Callable, Generator
from functools import reduce, wraps

from .base import SpotifyBase


def _chunks(lst: list, n: int, *, reverse: bool) -> Generator[list]:
    """
    Chunk list into length 'n' sublists.

    Parameters
    ----------
    lst
        list to chunk
    n
        length of chunks
    reverse
        reverse the order of chunks
    """
    rng = range(0, len(lst), n)

    if reverse:
        rng = rng[::-1]

    for i in rng:
        yield lst[i : i + n]


def _get_arg(position: int, name: str, args, kwargs):
    """Get argument from args or kwargs."""
    return kwargs.get(name, None) if len(args) <= position else args[position]


def _replace_arg(position: int, name: str, value, args, kwargs):
    if len(args) <= position:
        kwargs[name] = value
    else:
        args = (*args[:position], value, *args[position + 1 :])
    return args, kwargs


def chunked(  # noqa: C901
    arg_name: str,
    arg_pos: int,
    chunk_size: int,
    process: Callable,
    reverse: str | None = None,
    reverse_pos: int | None = None,
    chain: str | None = None,
    chain_pos: int | None = None,
) -> Callable:
    """
    Decorate a function to make multiple calls splitting a list argument.

    Optionally chain the return value of the previous request
    to the specified variable in the next request.

    Parameters
    ----------
    arg_name
        argument to chunk
    arg_pos
        index of argument in the argument list
    chunk_size
        size of chunk to send at a time
    process
        process response list
    reverse
        reverse order of chunks sent when argument defined
    reverse_pos
        position of the reverse argument
    chain
        variable to chain into the next request
    chain_pos
        position of the chain argument
    """

    def decorator(function: Callable) -> Callable:  # noqa: C901
        nonlocal arg_pos, reverse_pos, chain_pos
        arg_pos -= 1
        if reverse_pos is not None:
            reverse_pos -= 1
        if chain_pos is not None:
            chain_pos -= 1

        def replace(arg_val, chain_val, args, kwargs):
            args, kwargs = _replace_arg(arg_pos, arg_name, arg_val, args, kwargs)

            if chain is not None:
                args, kwargs = _replace_arg(chain_pos, chain, chain_val, args, kwargs)

            return args, kwargs

        async def async_wrapper(self: SpotifyBase, chunks, chain_val, args, kwargs):
            responses = []
            for chunk in chunks:
                args, kwargs = replace(chunk, chain_val, args, kwargs)
                chain_val = await function(self, *args, **kwargs)
                responses.append(chain_val)

            return process(responses)

        @wraps(function)
        def wrapper(self: SpotifyBase, *args, **kwargs):
            arg_val = _get_arg(arg_pos, arg_name, args, kwargs)
            if not self.chunked_on or arg_val is None:
                return function(self, *args, **kwargs)

            if chain is not None:
                chain_val = _get_arg(chain_pos, chain, args, kwargs)
            else:
                chain_val = None

            if reverse is not None:
                reverse_val = _get_arg(reverse_pos, reverse, args, kwargs)
                reverse_bool = reverse_val is not None
            else:
                reverse_bool = False

            chunks = _chunks(arg_val, chunk_size, reverse=reverse_bool)

            if self.is_async:
                return async_wrapper(self, chunks, chain_val, args, kwargs)

            responses = []
            for chunk in chunks:
                args, kwargs = replace(chunk, chain_val, args, kwargs)
                chain_val = function(self, *args, **kwargs)
                responses.append(chain_val)

            return process(responses)

        return wrapper

    return decorator


def join_lists(responses: list):
    """Join lists of models."""
    return reduce(operator.iadd, responses, [])


def return_none(_: list) -> None:
    """Return None always."""
    return


def return_last(responses: list):
    """Return last item of a list."""
    return responses[-1] if responses else None



from __future__ import annotations

from .serialise import Model, StrEnum


class ReleaseDatePrecision(StrEnum):
    """Precision of a release date."""

    year = "year"
    month = "month"
    day = "day"
    minute = "minute"


class Copyright(Model):
    """Copyright."""

    text: str
    type: str


class Followers(Model):
    """
    Followers.

    :attr:`href` is always ``None``.
    """

    href: None
    total: int


class Image(Model):
    """
    Image link and information.

    The Web API documentation reports that :attr:`height` and :attr:`width`
    can be ``None`` or not available in the response.
    """

    url: str
    height: int | None
    width: int | None


class Restrictions(Model):
    """Restrictions on relinked resource."""

    reason: str


class ResumePoint(Model):
    """Resume point."""

    fully_played: bool
    resume_position_ms: int
